CBWvar.PT_DEFEND=1
CBWvar.PT_ATTACK=2
CBWvar.PT_DEFATT=3
CBWvar.PT_STRONG=4
CBWvar.PT_STRONGDEF=5
CBWvar.PT_STRONGATT=6
CBWvar.PT_STRONGALL=7
CBWvar.PT_FLY=8
CBWvar.PT_KNUCKLESTOUCH=16
CBWvar.PT_GLIDE=32
CBWvar.PT_HAMMER=64
CBWvar.PT_TWINSPIN=128
CBWvar.PT_TAILBOUNCE=256

CBW_Battle.GetInteractionType = function(smo,tmo)
	//0 = No interaction. 1 = Bump. 2 = Nonlethal (unfinished). 3 = Full damage code
	if not(smo and smo.valid and smo.health and smo.player and not smo.player.spectator)
		or not(tmo and tmo.valid and tmo.health and tmo.player and not tmo.player.spectator)
		then return 0
	end
	if (gametype == GT_COMPETITION or gametype == GT_RACE)
		then return 0
	elseif (gametype == GT_TAG or gametype == GT_HIDEANDSEEK)
		then return 1
	elseif gametype == GT_MATCH
		then return 3
	elseif (gametype == GT_TEAMMATCH or gametype == GT_CTF) and smo.player.ctfteam != tmo.player.ctfteam
		then return 3
	elseif not multiplayer
		then return 0
	else return 0
	end
end

CBW_Battle.GetAttackPriorityType = function(mo)
	if not(mo and mo.valid and mo.health and mo.player) then return 0 end
	local player = mo.player
	local priority = 0
	//Spin attacks
	if (player.pflags&PF_JUMPED and not(player.pflags&PF_NOJUMPDAMAGE)) or player.pflags&PF_SPINNING then priority = $|CBWvar.PT_DEFATT end
	//Tails propellor attack
	if (player.charability == CA_FLY) and player.pflags&PF_THOKKED and not(player.pflags&PF_SHIELDABILITY) then priority = $|CBWvar.PT_DEFATT|CBWvar.PT_FLY end
	//Knuckles "fighter stance"
	if (player.charability == CA_GLIDEANDCLIMB) and P_IsObjectOnGround(mo) and not(player.pflags&PF_SPINNING) and not(mo.recoilangle) then priority = $|CBWvar.PT_KNUCKLESTOUCH|CBWvar.PT_DEFEND end
	//Gliding
	if player.pflags&PF_GLIDING then priority = $|CBWvar.PT_STRONGDEF|CBWvar.PT_GLIDE end
	//Amy's twinspin
	if (player.charability == CA_TWINSPIN and player.panim == PA_ABILITY) then priority = $|CBWvar.PT_STRONGALL|CBWvar.PT_TWINSPIN end
	//Ground hammer
	if (player.charability2 == CA2_MELEE and player.panim == PA_ABILITY2) then priority = $|CBWvar.PT_STRONGALL|CBWvar.PT_HAMMER end
	//Fang
	if (player.pflags&PF_BOUNCING) then priority = $|CBWvar.PT_STRONGATT|CBWvar.PT_TAILBOUNCE end
	//Metal
	if player.dashmode > 3*TICRATE and not(player.pflags&PF_STARTDASH) then priority = $|CBWvar.PT_STRONGDEF end
	return priority
end

CBW_Battle.GetCollideRelativeAngle = function(mo,collide)
	if not (mo and mo.valid and mo.health and mo.player and collide and collide.valid and collide.health)
		then return nil
	end
-- 	local collideangle = mo.player.drawangle - R_PointToAngle2(mo.x,mo.y,collide.x-collide.momx,collide.y-collide.momy)
	local collideangle = mo.angle - R_PointToAngle2(mo.x,mo.y,collide.x-collide.momx,collide.y-collide.momy)
	if collideangle >= 180*ANG1 then
		collideangle = $- 360*ANG1
	elseif collideangle <= -180*ANG1 then
		collideangle = $+360*ANG1
	end
	return collideangle
end

CBW_Battle.GetZCollideAngle = function(mo,collide)
	if not (mo and mo.valid and mo.health and collide and collide.valid and collide.health)	
		then return nil
	end
	local zdist = mo.z-collide.z
	local xydist = FixedHypot(mo.x-collide.x,mo.y-collide.y)
	local collideangle = R_PointToAngle2(0,0,xydist,zdist)
	return collideangle
end

CBW_Battle.GetAttackPriority = function(flags,xyangle,zangle)
	if flags&CBWvar.PT_KNUCKLESTOUCH and abs(xyangle) > ANGLE_45 then return 0 end //Forward sweetspot
	if flags&CBWvar.PT_FLY and (zangle > -ANGLE_45) then return 0 end //Upward sweetspot
	if flags&CBWvar.PT_TAILBOUNCE and (zangle < ANGLE_45) then return 0 end // Downward sweetspot
	if flags&CBWvar.PT_GLIDE and abs(zangle) > ANGLE_45 then // Forward and vertical center sweetspot
		if abs(xyangle) > ANGLE_45 then return CBWvar.PT_ATTACK
		else return CBWvar.PT_DEFATT end
	end
	if flags&CBWvar.PT_HAMMER and abs(xyangle) > ANGLE_45 then return CBWvar.PT_ATTACK end //Forward sweetspot
	if flags&CBWvar.PT_TWINSPIN and abs(xyangle) > ANGLE_45 and abs(zangle) < ANGLE_45 then return CBWvar.PT_ATTACK end //Blindspot to the sides and behind Amy
	//Specifications met. Return nonangled flags
	return flags & CBWvar.PT_STRONGALL
end

CBW_Battle.DoPlayerCollisionDamage = function(smo,tmo,spriority,tpriority)
	local sbias = 0 // -1 == send into disadvantage state. -2 == take damage
	local tbias = 0
	local satk = 0 // 1 for normal attacks, 2 for strong attacks, 3 for very strong attacks
	local tatk = 0
	local sdef = 0 // 1 for normal defense, 2 for strong defense, 3 for very strong attacks
	local tdef = 0
	if spriority&CBWvar.PT_DEFEND then sdef = $+1 end
	if tpriority&CBWvar.PT_DEFEND then tdef = $+1 end
	if spriority&CBWvar.PT_ATTACK then satk = $+1 end
	if tpriority&CBWvar.PT_ATTACK then tatk = $+1 end
	if spriority&CBWvar.PT_STRONG then
		sdef = $+2
		satk = $+2
	end
	if tpriority&CBWvar.PT_STRONG then
		tdef = $+2 
		tatk = $+2
	end
	sbias = sdef-tatk
	tbias = tdef-satk
	
	local spower = (smo.player.powers[pw_super] or smo.player.powers[pw_invulnerability])
	local tpower = (tmo.player.powers[pw_super] or tmo.player.powers[pw_invulnerability])
	local stagit = smo.player.pflags&PF_TAGIT
	local ttagit = tmo.player.pflags&PF_TAGIT
	local sinvuln = smo.player.powers[pw_flashing]
	local tinvuln = tmo.player.powers[pw_flashing]
	if spower and not(tpower or tinvuln) then P_DamageMobj(tmo,smo,smo,0) return 3
	elseif tpower and not(spower or sinvuln) then P_DamageMobj(smo,tmo,tmo,0) return -3
	/*
	elseif stagit and not(ttagit) then
		if not(CBW_Battle.TagNakedHit(tmo,smo,smo,true)) then
			P_DamageMobj(tmo,smo,smo,0)
		end
	elseif ttagit and not(stagit) then
		if not(CBW_Battle.TagNakedHit(tmo,smo,smo,true)) then
			P_DamageMobj(smo,tmo,tmo,0)
		end
	*/
	else
		if sbias < 0 and not(sinvuln or spower) then P_DamageMobj(smo,tmo,tmo,0) end
		if tbias < 0 and not(tinvuln or tinvuln) then P_DamageMobj(tmo,smo,smo,0) end
	end
	return sbias-tbias //Used for recoil and knockback distribution
end

CBW_Battle.UpdateRecoilState = function(mo)
	if not(mo and mo.valid and mo.health and mo.player) then return nil //Object invalid
	elseif mo.player.powers[pw_nocontrol] > 0 and mo.recoilangle != nil and mo.recoilthrust != nil then
		if P_PlayerInPain(mo.player) then
			mo.recoilangle = nil
			mo.recoilthrust = nil
			return false
		end
		if P_IsObjectOnGround(mo) then //Grounded
			if not(mo.player.charability == CA_GLIDEANDCLIMB) then
				mo.state = S_PLAY_SKID
				mo.player.panim = 0
				mo.frame = 0
			else
				mo.state = S_PLAY_GLIDE_LANDING
				mo.player.powers[pw_nocontrol] = $-1
			end
			if mo.player.skidtime == 0 and mo.recoilthrust > mo.scale*5 then
				S_StartSound(mo, sfx_skid)
			end
			mo.player.skidtime = mo.player.powers[pw_nocontrol]
			//Apply recoil thrust
			P_InstaThrust(mo,mo.recoilangle,mo.recoilthrust)
			//Add friction
			mo.recoilthrust = ($+FixedMul($,FixedMul(mo.friction,mo.movefactor)))/2
		else //Aerial
			//Keep our variables up with current trajectory
			mo.recoilthrust = FixedHypot(mo.momx,mo.momy)
			mo.recoilangle = R_PointToAngle2(0,0,mo.momx,mo.momy)
			mo.player.panim = PA_FALL
			mo.state = S_PLAY_FALL
		end
		mo.player.drawangle = mo.recoilangle+ANGLE_180
		return true
	else //Not in disadvantage state, reset variables
		mo.recoilangle = nil
		mo.recoilthrust = nil
		return false
	end
end

CBW_Battle.DoPlayerInteract = function(smo,tmo)
	if not(smo and tmo and smo.valid and tmo.valid and smo.health and tmo.health)
		or (smo.pushed == nil or tmo.pushed == nil)
		or ((smo.pushed_last == smo.pushed) or (tmo.pushed_last == tmo.pushed))
		then return false
	end
	local collisiontype = CBW_Battle.GetInteractionType(smo,tmo)
	local splr = smo.player
	local tplr = tmo.player
	
	//No more homing
	splr.homing = 0
	tplr.homing = 0
	
	//Get thrust
	local sspd = FixedHypot(smo.momx,smo.momy)
	local tspd = FixedHypot(tmo.momx,tmo.momy)
	local szspd = abs(smo.momz)
	local tzspd = abs(tmo.momz)
	local sangle = R_PointToAngle2(tmo.x-tmo.momx,tmo.y-tmo.momy,smo.x-smo.momx,smo.y-smo.momy)
	local tangle = R_PointToAngle2(smo.x-smo.momx,smo.y-smo.momy,tmo.x-tmo.momx,tmo.y-tmo.momy)
	local soundthrust = min(tmo.scale,smo.scale)
	local sthrust = FixedMul(tspd*2/3,tmo.scale-smo.scale/2)+smo.scale
	local sthrust2 = FixedMul(sspd+sspd/3,tmo.scale-smo.scale/2)+smo.scale
	local tthrust = FixedMul(sspd*2/3,smo.scale-tmo.scale/2)+tmo.scale
	local tthrust2 = FixedMul(tspd+tspd/3,smo.scale-tmo.scale/2)+tmo.scale
	local smomz = smo.momz
	local tmomz = tmo.momz

	
	//Get collision angle
	local tcollideangle = CBW_Battle.GetCollideRelativeAngle(tmo,smo)
	local scollideangle = CBW_Battle.GetCollideRelativeAngle(smo,tmo)
	
	local tzcollideangle = CBW_Battle.GetZCollideAngle(tmo,smo)
	local szcollideangle = CBW_Battle.GetZCollideAngle(smo,tmo)
	local tpriority = 0
	local spriority = 0
	local tprioritytype = 0
	local sprioritytype = 0
	
	local bias = 0
	//Collision info
	if collisiontype <= 1 and G_TagGametype() and not(leveltime < 30*TICRATE) then
		if tplr.pflags&PF_TAGIT and not(splr.pflags&PF_TAGIT) and CBW_Battle.PlayerCanBeDamaged(splr) then
			splr.pflags = $|PF_TAGIT
			P_DoPlayerPain(smo.player, tmo, tmo)
			P_AddPlayerScore(tmo.player,100)
			if multiplayer and not(splitscreen) then print(CBW_Battle.CustomHurtMessage(smo.player,tmo,"tagging hand")) end
		elseif splr.pflags&PF_TAGIT and not(tplr.pflags&PF_TAGIT) and CBW_Battle.PlayerCanBeDamaged(tplr) then
			tplr.pflags = $|PF_TAGIT
			P_DoPlayerPain(tmo.player, smo, smo)
			P_AddPlayerScore(smo.player,100)
			if multiplayer and not(splitscreen) then print(CBW_Battle.CustomHurtMessage(tmo.player,smo,"tagging hand")) end
		end
	elseif collisiontype > 1 then
		//Get collision prioritytype
		tprioritytype = CBW_Battle.GetAttackPriorityType(tmo)
		sprioritytype = CBW_Battle.GetAttackPriorityType(smo)
		
		//Update attack priority
		tpriority = CBW_Battle.GetAttackPriority(tprioritytype,tcollideangle,tzcollideangle)
		spriority = CBW_Battle.GetAttackPriority(sprioritytype,scollideangle,szcollideangle)
		
		//Do collision damage
		bias = CBW_Battle.DoPlayerCollisionDamage(smo,tmo,spriority,tpriority)
	end
	
	local spain = (splr.powers[pw_nocontrol] > 0 or splr.panim == PA_PAIN or smo.health < 1)
	local tpain = (tplr.powers[pw_nocontrol] > 0 or tplr.panim == PA_PAIN or tmo.health < 1)
	
	//Apply Thrust
	if smo.health then
		P_Thrust(smo,sangle,max(tmo.scale,sthrust))
		if not spain then P_Thrust(smo,tangle+ANGLE_180,max(tmo.scale,sthrust2)) end
	end
	if tmo.health then
		P_Thrust(tmo,tangle,max(smo.scale,tthrust))
		if not tpain then P_Thrust(tmo,sangle+ANGLE_180,max(smo.scale,tthrust2)) end
	end
	
	//Get and apply zthrust
	if not(spain) then smo.momz = FixedMul(tmomz-smomz,tmo.scale*2-smo.scale) end
	if not(tpain) then tmo.momz = FixedMul(smomz-tmomz,smo.scale*2-tmo.scale) end
	
	//Knuckles interaction
	if smo.health and not(spain)
		and (splr.pflags&PF_GLIDING or splr.climbing 
			or (splr.charability == CA_GLIDEANDCLIMB) and collisiontype > 1 and P_IsObjectOnGround(smo) and not(splr.pflags&PF_SPINNING)
		) then
		splr.pflags = $&~(PF_GLIDING|PF_JUMPED)
		if not(P_IsObjectOnGround(smo)) or splr.climbing then
			splr.panim = PA_FALL
			splr.mo.state = S_PLAY_FALL
			splr.climbing = 0
			splr.lastsidehit = -1
			splr.lastlinehit = -1
		elseif (tpriority&CBWvar.PT_ATTACK)
			splr.mo.state = S_PLAY_GLIDE_LANDING
			smo.momz = 0
		end
	end
	if tmo.health and not(tpain)
		and (tplr.pflags&PF_GLIDING or tplr.climbing 
			or (tplr.charability == CA_GLIDEANDCLIMB) and collisiontype > 1 and P_IsObjectOnGround(tmo) and not(tplr.pflags&PF_SPINNING)
		) then
		tplr.pflags = $&~(PF_GLIDING|PF_JUMPED)
		if not(P_IsObjectOnGround(tmo)) or tplr.climbing then
			tplr.panim = PA_FALL
			tplr.mo.state = S_PLAY_FALL
			tplr.climbing = 0
			tplr.lastsidehit = -1
			tplr.lastlinehit = -1
		elseif (spriority&CBWvar.PT_ATTACK)
			tplr.mo.state = S_PLAY_GLIDE_LANDING
		end
	end
	
	//Do recoil state
	if collisiontype > 1 then
		if not(spain) and (bias == -1 or FixedHypot(sspd,szspd) < FixedHypot(tspd,tzspd)) 
		and not(sprioritytype&(CBWvar.PT_HAMMER|CBWvar.PT_TWINSPIN))
		then
			splr.powers[pw_nocontrol] = sthrust/smo.scale*3
			splr.panim = PA_FALL
			smo.state = S_PLAY_FALL
			smo.recoilangle = sangle
			smo.recoilthrust = sthrust
			splr.pflags = $&~(PF_GLIDING|PF_JUMPED|PF_BOUNCING|PF_THOKKED)
		end
		if not(tpain) and (bias == 1 or FixedHypot(sspd,szspd) > FixedHypot(tspd,tzspd))
		and not(tprioritytype&(CBWvar.PT_HAMMER|CBWvar.PT_TWINSPIN))
		then
			tplr.powers[pw_nocontrol] = tthrust/tmo.scale*3
			tplr.panim = PA_FALL
			tmo.state = S_PLAY_FALL
			tmo.recoilangle = tangle
			tmo.recoilthrust = tthrust
			tplr.pflags = $&~(PF_GLIDING|PF_JUMPED|PF_BOUNCING|PF_THOKKED)
		end
	end
	//Do Sound
	local defend = (((tpriority|spriority)&CBWvar.PT_DEFEND)>0)
	local attack = (((tpriority|spriority)&CBWvar.PT_ATTACK)>0)
	local strong = (((tpriority|spriority)&CBWvar.PT_STRONG)>0)
	local damage = (tpain or spain)
	if not(damage or defend or attack or strong) or (damage and not strong) then
		S_StartSound(smo,sfx_s3k5d)
	elseif not strong then
		S_StartSound(smo,sfx_s3k49)
	else
		S_StartSound(smo,sfx_s3k5f)
	end

	//Player interactions completed
	return true
end

CBW_Battle.PlayerTouch = function(smo,tmo)
	if not(smo and smo.valid and tmo and tmo.valid and smo.player and tmo.player)
		or not(smo.health and tmo.health)
		or (smo.player.exiting or tmo.player.exiting)
		then return true end
	
	if not(CBW_Battle.GetInteractionType(smo,tmo)) then
		if not(tmo.player.powers[pw_carry] or smo.player.powers[pw_carry] or P_PlayerInPain(tmo.player) or P_PlayerInPain(smo.player)) then		
			local sf = P_MobjFlip(smo)
			local tf = P_MobjFlip(tmo)
			if tmo.player.pflags&PF_CANCARRY/* and tmo.momz*tf > 0*/ and tmo.z*tf > (smo.z+smo.height/8*7)*sf then
				tmo.flags = $&~MF_SPECIAL			
				smo.flags = $&~MF_SPECIAL
			elseif smo.player.pflags&PF_CANCARRY /*and smo.momz*sf > 0*/ and smo.z*sf > (tmo.z+tmo.height/8*7)*tf then
				tmo.flags = $&~MF_SPECIAL			
				smo.flags = $&~MF_SPECIAL
			end
		end
	return true end //Don't bother trying to collide if nothing's going to come of it
	
	smo.pushed = tmo //'mo.pushed' will be referenced in ThinkFrame
	tmo.pushed = smo
	smo.flags = $&~MF_SPECIAL
	tmo.flags = $&~MF_SPECIAL
	tmo.player.pflags = $&~PF_CANCARRY
	smo.player.pflags = $&~PF_CANCARRY
	return true
end

CBW_Battle.UpdateCollisionHistory = function(pmo)
	if pmo.pushtics == nil then pmo.pushtics = 0 end //!! This shouldn't be necessary. Investigate later.
	if pmo.pushed then
		if pmo.pushtics == 0 then
			pmo.pushtics = 1
		else
			pmo.pushtics = $+1
			if pmo.pushtics > 15 then
				pmo.pushed_last = nil
				pmo.pushed = nil
				pmo.pushtics = 0
			end
		end
	else
		pmo.pushtics = 0
	end
	if pmo.pushed and pmo.pushed.valid and pmo.pushed.pushed == pmo and not(pmo.pushtics) then //Update other player object's pushed stats
		pmo.pushed.flags = $|MF_SPECIAL
		pmo.pushed.pushed_last = pmo
		pmo.pushed.pushed = nil
	end
	pmo.flags = $|MF_SPECIAL
	pmo.pushed_last = pmo.pushed
	pmo.pushed = nil
end